import pandas as pd
from dios.options import *
from dios.lib import *
from dios.errors import *


class CastPolicy:
    force = 'force'
    save = 'save'
    never = 'never'


CAST_POLICIES = get_storage_class_values(CastPolicy)


class __Itype:
    def __init__(self):
        raise RuntimeError("DatetimeItype does not allow instances of itself.")


class DatetimeItype(__Itype):
    name = 'datetime'
    repr = 'DatetimeItype'
    unique = True
    subtypes = (pd.DatetimeIndex,)


class IntegerItype(__Itype):
    name = 'integer'
    repr = 'IntegerItype'
    unique = True
    subtypes = (pd.RangeIndex, pd.Int64Index, pd.UInt64Index, int)


class FloatItype(__Itype):
    name = 'float'
    repr = 'FloatItype'
    subtypes = (pd.Float64Index, float)
    unique = True


# class MultiItype(__Itype):
#     name = "multi"
#     repr = 'MultiItype'
#     subtypes = (pd.MultiIndex, )
#     unique = ??


class NumericItype(__Itype):
    name = "numeric"
    repr = 'NumericItype'
    _subitypes = (IntegerItype, FloatItype)
    subtypes = (_subitypes + IntegerItype.subtypes + FloatItype.subtypes)
    unique = False


class MixedItype(__Itype):
    name = "mixed"
    repr = 'MixedItype'
    unique = False
    _subitypes = (DatetimeItype, IntegerItype, FloatItype, NumericItype)
    _otheritypes = (pd.CategoricalIndex, pd.IntervalIndex, pd.PeriodIndex, pd.TimedeltaIndex, pd.Index)
    subtypes = (_subitypes + _otheritypes + DatetimeItype.subtypes + NumericItype.subtypes)


def is_itype(obj, itype):
    """ Check if obj is a instance of the given itype or its str-alias was given"""

    # todo: iter through itype as it could be a tuple, if called like ``is_itype(o, (t1,t2))``

    # user gave a Itype, like ``DatetimeItype``
    if type(obj) == type and issubclass(obj, itype):
        return True

    # user gave a string, like 'datetime'
    if isinstance(obj, str) and obj == itype.name:
        return True

    return False


def is_itype_subtype(obj, itype):
    """ Check if obj is a subclass or a instance of a subclass of the given itype"""

    # user gave a subtype, like ``pd.DatetimeIndex``
    if type(obj) == type and issubclass(obj, itype.subtypes):
        return True

    # user gave a instance of a subtype, like ``pd.Series(..).index``
    if isinstance(obj, itype.subtypes):
        return True

    return False


def is_itype_like(obj, itype):
    """ Check if obj is a subclass or a instance of the given itype or any of its subtypes"""
    return is_itype(obj, itype) or is_itype_subtype(obj, itype)


def get_itype(obj):
    """
    Return the according Itype, by any of any possible user input, like
      - "datetime"
      - DatetimeItype
      - pd.Series(...).index
      - pd.DatetimeIndex
    and return the according Itype
    """
    if type(obj) == type and issubclass(obj, __Itype):
        return obj

    # check if it is the actual type, not a subtype
    types = [DatetimeItype, IntegerItype, FloatItype, NumericItype, MixedItype]
    for t in types:
        if is_itype(obj, t):
            return t

    for t in types:
        if is_itype_subtype(obj, t):
            return t

    raise ValueError(f"{obj} is not a itype, nor any known subtype of a itype, nor a itype string alias")


def itype_eq(a, b):
    return is_itype(a, b)


def itype_lt(a, b):
    return is_itype_subtype(a, b)


def itype_le(a, b):
    return is_itype_like(a, b)


def cast_to_itype(series, itype, policy='lossless', err='raise', inplace=False):
    """ Cast a series (more explicit the type of the index) to fit the itype of a dios.

    Return the casted series if successful, None otherwise.

    Note:
        This is very basic number-casting, so in most cases, information from
        the old index will be lost after the cast.
    """

    if policy not in CAST_POLICIES:
        raise ValueError(f"policy={policy}")
    if err not in ['raise', 'ignore']:
        raise ValueError(f"err={err}")
    if not inplace:
        series = series.copy()
    series.itype = get_itype(series.index)

    # up-cast issn't necessary because a dios with a higher
    # itype always can take lower itypes.
    # series can have dt/int/float/mixed
    # dt    -> dt           -> mixed
    # int   -> int   -> num -> mixed
    # float -> float -> num -> mixed
    # mixed                 -> mixed
    if itype_le(series.itype, itype):  # a <= b
        return series

    e = f"A series index of type `{type(series.index)}` cannot be casted to Itype {itype.name}"

    # cast any -> dt always fail.
    if is_itype(itype, DatetimeItype):
        pass
    else:
        e += f", as forbidden by the cast-policy `{policy}`."

    if policy == CAST_POLICIES[CastPolicy.never]:
        pass

    elif policy == CAST_POLICIES[CastPolicy.force]:
        # cast any (dt/float/mixed) -> int
        # cast any (dt/float/mixed) -> num
        if is_itype(itype, IntegerItype) or is_itype(itype, NumericItype):  # a == b or a == c
            series.index = pd.RangeIndex(len(series))
            return series
        # cast any (dt/int/mixed) -> float
        if is_itype(itype, FloatItype):  # a == b
            series.index = pd.Float64Index(range(len(series)))
            return series

    elif policy == CAST_POLICIES[CastPolicy.save]:
        # cast int   -> float
        if is_itype(itype, IntegerItype) and is_itype(series.itype, FloatItype):  # a == b and c == d
            series.index = series.index.astype(float)
            return series
        # cast float -> int, maybe if unique
        if is_itype(itype, FloatItype) and is_itype(series.itype, IntegerItype):  # a == b and c == d
            series.index = series.index.astype(int)
            if series.index.is_unique:
                return series
            e = f"The cast with policy {policy} from series index type `{type(series.index)}` to " \
                f"itype {itype.name} resulted in a non-unique index."
        # cast mixed -> int/float always fail

    if err == 'raise':
        raise ItypeCastError(e)
    else:
        return None


def throw_MixedItype_err_or_warn(itype):
    msg = f"'{itype.repr}'. Using a non-unique Itype for a DictionaryOfSeries, is not recommend. " \
          f"As soon as series \nwith different index types are inserted, selecting and slicing data will " \
          f"almost always fail. You are hereby warned!"
    if dios_options[OptsFields.mixed_itype_warn_policy] in ['ignore', Opts.itype_ignore]:
        pass
    elif dios_options[OptsFields.mixed_itype_warn_policy] in ['error', 'err', Opts.itype_err]:
        raise ItypeCastError(msg)
    else:
        warnings.warn(msg, ItypeWarning)
    return
